#    Healthcheck Bot
#    Copyright (C) 2018 Dmitry Berezovsky
#    
#    HealthcheckBot is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#    
#    HealthcheckBot is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#    
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

from datetime import datetime, time

import requests

from healthcheckbot.common import validators
from healthcheckbot.common.model import WatcherModule, ParameterDef, ValidationReporter


class SystemTimeWatcher(WatcherModule):

    def __init__(self, application):
        super().__init__(application)
        self.error_when_midnight = False

    def obtain_state(self, trigger) -> object:
        current_time = datetime.now()
        return current_time

    def serialize_state(self, state: datetime) -> [dict, None]:
        return {
            "time": state.isoformat()
        }

    def do_assertions(self, state: datetime, reporter: ValidationReporter):
        if self.error_when_midnight:
            if state.time() == time(0, 0):
                reporter.error('its_midnight')

    PARAMS = (
        ParameterDef('error_when_midnight', validators=(validators.boolean,)),
    )


class HttpRequest(WatcherModule):

    def __init__(self, application):
        super().__init__(application)
        self.url = None
        self.headers = None
        self.auth_user = None
        self.auth_password = None
        self.payload = None
        self.method = 'GET'
        self.timeout = 4
        self.assert_status = None
        self.assert_response_time = None

    def obtain_state(self, trigger):
        return requests.request(self.method, self.url, data=self.payload, headers=self.headers,
                                auth=self.basic_auth,
                                timeout=self.timeout)

    @property
    def basic_auth(self):
        return None if self.auth_user is None and self.auth_password is None else (self.auth_user, self.auth_password)

    def serialize_state(self, state: requests.Response):
        return dict(
            status_code=state.status_code,
            response_time=state.elapsed.total_seconds(),
            url=state.url,
            redirect_history=[dict(status_code=x.status_code, url=x.url) for x in state.history]
        )

    def do_assertions(self, state: requests.Response, reporter: ValidationReporter):
        if self.assert_status is not None and state.status_code != self.assert_status:
            reporter.error('status_code',
                           'Expected HTTP status {} but got {}'.format(self.assert_status, state.status_code))
        if self.assert_response_time is not None and state.elapsed.total_seconds() > self.assert_response_time:
            reporter.error('response_time',
                           'Expected HTTP response time must be '
                           '< {} but actual is {}'.format(self.assert_response_time, state.elapsed.total_seconds()))

    PARAMS = (
        ParameterDef('url', is_required=True, validators=(validators.string,)),
        ParameterDef('method', validators=(validators.string,)),
        ParameterDef('auth_user', validators=(validators.string,)),
        ParameterDef('auth_password', validators=(validators.string,)),
        ParameterDef('headers', validators=(validators.dict_of_strings,)),
        ParameterDef('timeout', validators=(validators.integer,)),
        ParameterDef('payload'),
        ParameterDef('assert_status'),
        ParameterDef('assert_response_time', validators=(validators.number,)),
    )



#    Healthcheck Bot
#    Copyright (C) 2018 Dmitry Berezovsky
#    
#    HealthcheckBot is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#    
#    HealthcheckBot is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#    
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging

import sys
from typing import Callable, List

import os
import yaml

from healthcheckbot.common.core import ApplicationManager
from healthcheckbot.common.error import ConfigValidationError
from healthcheckbot.common.evaluator import simple_env_evaluator
from healthcheckbot.common.model import Module, WatcherModule
from healthcheckbot.common.utils import EvaluatingConfigWrapper

logger = logging.getLogger('Bootstrap')

__all__ = ['read_config', 'bootstrap', 'bootstrap_from_cli']


def read_config(config_file) -> dict:
    if isinstance(config_file, str):
        config_file = open(config_file, mode='r')
    try:
        return yaml.load(config_file)
    finally:
        if hasattr(config_file, 'close'):
            config_file.close()


def bootstrap(config: dict) -> ApplicationManager:
    application = bootstrap_environment(config)
    return bootstrap_from_cli(config, application)


def bootstrap_from_cli(config: dict, application: ApplicationManager) -> ApplicationManager:
    # Instantiate components
    instantiate_outputs(config, application)
    return application


def bootstrap_cli(config: dict) -> ApplicationManager:
    return bootstrap_environment(config, True)


def bootstrap_environment(config: dict, enable_cli=False) -> ApplicationManager:
    application = new_application()
    logger.info('Reading config file')
    config = add_expression_evaluator(config)
    save_instance_config(config, application)
    load_context_path(application)
    return application


def new_application(enable_cli=False):
    application = ApplicationManager()
    application.get_instance_settings().enable_cli = enable_cli
    return application


def add_expression_evaluator(config_section: dict) -> EvaluatingConfigWrapper:
    return EvaluatingConfigWrapper(config_section, simple_env_evaluator)


def save_instance_config(config: dict, application: ApplicationManager):
    if 'app' in config:
        app_config = add_expression_evaluator(config.get('app', {}))
        settings = application.get_instance_settings()
        settings.id = app_config.get('id', 'healthcheckbot0')
        # Context Path
        for path in app_config.get('classpath', []):
            if isinstance(path, str):
                if not os.path.exists(path):
                    raise ConfigValidationError('app/classpath', 'path ' + path + ' doesn\'t exist')
                settings.context_path.append(path)
            else:
                raise ConfigValidationError('app/classpath', 'Context path must be the list of strings')


def load_context_path(application: ApplicationManager):
    context_path = application.get_instance_settings().context_path
    for p in context_path:
        sys.path.append(p)


def instantiate_modules_for_section(section_name: str, section_config: dict, application: ApplicationManager,
                                    cb: Callable[[ApplicationManager, Module, dict], None] = None,
                                    instance_name_prefix='') -> List[Module]:
    result = []
    for name, module_def in section_config.items():
        if not (isinstance(module_def, dict) and 'provider' in module_def):
            raise ConfigValidationError('/'.join((section_name, name)),
                                        'Should be dictionary containing mandatory key "provider"')
        module_def = add_expression_evaluator(module_def)
        instance = application.register_module_instance(instance_name_prefix + name, module_def.get('provider'))
        try:
            # Parse and set parameters
            for param_def in instance.PARAMS:
                if param_def.name in module_def:
                    val = module_def.get(param_def.name)
                    if param_def.parser is not None:
                        val = param_def.parser.parse(val, application, '/'.join((section_name, name, param_def.name)))
                    val = param_def.sanitize(val)
                    param_def.validate(val)
                    setattr(instance, param_def.name, val)
                elif param_def.is_required:
                    raise ConfigValidationError('/'.join((section_name, name, param_def.name)),
                                                'Parameter {} is required'.format(param_def.name))
            # Run validation of the module device
            instance.validate()
            if cb is not None:
                cb(application, instance, module_def)
            instance.on_configured()
            logger.info('Configured module {}({}))'.format(name, instance.__class__.__name__))
            result.append(instance)
        except Exception as e:
            raise ConfigValidationError('/'.join((section_name, name)), "Invalid module configuration: " + str(e), e)
    return result


def _register_watcher_module(application: ApplicationManager, watcher: WatcherModule, module_def: dict):
    for trigger_name in module_def.get('triggers', tuple()):
        trigger = application.get_trigger_by_name(trigger_name)
        if trigger is None:
            raise ConfigValidationError('triggers', "Trigger \'{}\' doesn't exist".format(trigger_name))
        trigger.register_watcher(watcher)
    assertions = instantiate_modules_for_section('watchers/{}/custom_assertions'.format(watcher.name),
                                                 module_def.get('custom_assertions', dict()), application,
                                                 instance_name_prefix=watcher.name + '__')
    watcher.custom_assertions = assertions


def instantiate_outputs(config: dict, application: ApplicationManager):
    section_config = config.get('outputs', {})
    instantiate_modules_for_section('outputs', section_config, application)
    section_config = config.get('triggers', {})
    instantiate_modules_for_section('triggers', section_config, application)
    section_config = config.get('watchers', {})
    instantiate_modules_for_section('watchers', section_config, application, _register_watcher_module)



#!/usr/bin/env python3

#    Healthcheck Bot
#    Copyright (C) 2018 Dmitry Berezovsky
#    
#    HealthcheckBot is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#    
#    HealthcheckBot is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#    
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

import argparse
import logging

import gc
import os

from healthcheckbot import app
from healthcheckbot.common import bootstrap
from healthcheckbot.common.model import CliExtension
from healthcheckbot.common.utils import CLI

DEFAULT_CONFIG_LOCATIONS = [
    './config.yaml',
    './config.yml',
]

supplementary_parser = argparse.ArgumentParser(add_help=False)
supplementary_parser.add_argument("-c", "--config", dest='config', type=argparse.FileType('r'), required=False,
                                  help="File containing HealthcheckBot configuration in yaml or JSON format")
supplementary_parser.add_argument("-v", "--verbose", dest="verbose", action='store_true', required=False,
                                  help="If set more verbose output will used",
                                  default=False)

parser = argparse.ArgumentParser(prog='healthcheckbot', add_help=True,
                                 description='Advanced health and status checks for your software',
                                 formatter_class=argparse.RawDescriptionHelpFormatter, epilog="""
--------------------------------------------------------------
Healthcheck Bot Copyright (C) 2019 Dmitry Berezovsky
This program comes with ABSOLUTELY NO WARRANTY;
This is free software, and you are welcome to redistribute it
under certain conditions;
--------------------------------------------------------------
"""
                                 )

parser.add_argument("-c", "--config", dest='config', type=argparse.FileType('r'), required=False,
                    help="File containing HealthcheckBot configuration in yaml or JSON format")

parser.add_argument("-v", "--verbose", dest="verbose", action='store_true', required=False,
                    help="If set more verbose output will used",
                    default=False)

module_parser = parser.add_subparsers(title="module",
                                      dest="module",
                                      description='Use \"<MODULE_NAME> -h\" to get information '
                                                  'about command available for particular module')

_config = None
namespace_parsers = {}


class ApplicationRunCLIExtension(CliExtension):
    COMMAND_NAME = 'run'
    COMMAND_DESCRIPTION = 'Run Healthcheck Bot in foreground'

    def handle(self, args):
        global _config
        bootstrapped_app = bootstrap.bootstrap_from_cli(_config, self.get_application_manager())
        app.run_application(_config, bootstrapped_app)


class VerifyConfigCLIExtension(CliExtension):
    COMMAND_NAME = 'verify'
    COMMAND_DESCRIPTION = 'Verify configuration file end exit'

    def handle(self, args):
        global _config
        try:
            bootstrapped_app = bootstrap.bootstrap_from_cli(_config, self.get_application_manager())
            CLI.print_info("Config file is valid")
        except Exception as e:
            CLI.print_error("Config file is invalid")
            CLI.print_error(e)
            exit(1)


def get_default_config_path():
    for config_path in DEFAULT_CONFIG_LOCATIONS:
        if os.path.isfile(config_path):
            return config_path
    return None


def read_config_from_arguments() -> dict:
    known, unknown = supplementary_parser.parse_known_args()
    CLI.verbose_mode = known.verbose
    if known.config is None:
        config_path = get_default_config_path()
    else:
        config_path = known.config.name
        known.config.close()
    if config_path:
        CLI.print_info("Loaded config: " + os.path.realpath(config_path))
        CLI.print_info("")
        return bootstrap.read_config(config_path)
    else:
        CLI.print_error('Unable to load config file. Check -c/--config option.')
        exit(1)


def configure_subparser_for_cli_extension(ext, parser, application):
    ext.setup_parser(parser)
    ext_instance = ext(parser, application)
    parser.set_defaults(handler=ext_instance)


def main():
    global _config
    logging.basicConfig(level=logging.DEBUG)
    _config = read_config_from_arguments()
    application = bootstrap.bootstrap_cli(_config)
    # Register root commands
    root_commands = (ApplicationRunCLIExtension, VerifyConfigCLIExtension)
    # Process root commands
    for cmd in root_commands:
        cmd_subparser = module_parser.add_parser(cmd.COMMAND_NAME, help=cmd.COMMAND_DESCRIPTION)
        configure_subparser_for_cli_extension(cmd, cmd_subparser, application)

    # Process CLI extensions
    for namespace, ext in application.cli_extensions:
        if namespace not in namespace_parsers.keys():
            p = module_parser.add_parser(namespace)
            namespace_parsers[namespace] = p.add_subparsers(title='module_cmd', dest='module_cmd')
        ext_parser = namespace_parsers[namespace]
        if ext.COMMAND_NAME is None:
            raise Exception(
                "Invalid CLI Extension: {}. COMMAND_NAME needs to be defined".format(ext.__class__.__name__))
        ext_subparser = ext_parser.add_parser(ext.COMMAND_NAME,
                                              help=ext.COMMAND_DESCRIPTION)
        configure_subparser_for_cli_extension(ext, ext_subparser, application)

    args = parser.parse_args()
    gc.collect()
    if hasattr(args, "handler"):
        args.handler.handle(args)


if __name__ == '__main__':
    main()



#    Healthcheck Bot
#    Copyright (C) 2018 Dmitry Berezovsky
#    
#    HealthcheckBot is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#    
#    HealthcheckBot is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#    
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

from os import path
from setuptools import setup, find_packages

from healthcheckbot.version import version

src_dir = path.abspath(path.dirname(__file__))
root_dir = path.join(src_dir, '..')

# Get the long description from the README file
readme_file = path.join(root_dir, 'README.MD')
try:
    from m2r import parse_from_file

    long_description = parse_from_file(readme_file)
except ImportError:
    # m2r may not be installed in user environment
    with open(readme_file) as f:
        long_description = f.read()

setup(
    name='healthcheckbot',
    # Semantic versioning should be used:
    # https://packaging.python.org/distributing/?highlight=entry_points#semantic-versioning-preferred
    version=version,
    description='Robust way to setup external health checks for your software',
    long_description=long_description,
    url='https://githib.com/Logicify/healthcheckbot',
    keywords='python healthchecks assertion validate health server healthcheck',

    # Author
    author='Dmitry Berezovsky',

    # License
    license='GPLv3',

    # Technical meta
    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 3 - Alpha',

        # Who your project is intended for
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Build Tools',

        # License (should match "license" above)
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        # Python versions support
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
    ],

    # Structure
    packages=find_packages(include=['healthcheckbot', 'healthcheckbot.*']),
    # py_modules=["app", 'cli', 'daemonize'],

    install_requires=[
        'PyYAML==4.2b1',
        'typing>=3.6',
        'requests==2.21.0',
        'beautifulsoup4==4.4.0',
        'graypy==0.3.2',
        'daemons==1.3.0',

    ],

    # Extra dependencies might be installed with:
    # pip install -e .[dev,test]
    extras_require={
        'dev': [],
        'test': [],
    },

    package_data={
        'examples': [path.join(root_dir, 'examples')],
    },

    test_suite='nose2.collector.collector',
    tests_require=[
        'nose2==0.8.0',
    ],
    entry_points={
        'console_scripts': [
            'healthcheckbot=healthcheckbot.cli:main',
        ],
    }
)


#    Healthcheck Bot
#    Copyright (C) 2018 Dmitry Berezovsky
#    
#    HealthcheckBot is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#    
#    HealthcheckBot is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#    
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging
import time
from typing import List, Tuple, Type, NamedTuple, Dict

from healthcheckbot.common.error import InvalidModuleError, WatcherRuntimeError, OutputRuntimeError
from healthcheckbot.common.model import CliExtension, Module, TriggerModule, OutputModule, WatcherModule, \
    ValidationReporter, WatcherResult, LoopModuleMixin, WatcherAssert


class ModuleType(NamedTuple):
    bucket: dict
    name: str
    base_class: Type


class InstanceSettings:
    def __init__(self):
        self.id = None
        self.enable_cli = False
        self.context_path = []


class ApplicationManager:

    def __init__(self) -> None:
        self.__instance_settings = InstanceSettings()
        self.__terminating = False
        self.cli_extensions = []  # type: List[Tuple[str, CliExtension]]
        self.__logger = logging.getLogger('ApplicationManager')
        self.__triggers = {}  # type: Dict[TriggerModule]
        self.__outputs = {}  # type: Dict[OutputModule]
        self.__watchers = {}  # type: Dict[WatcherModule]
        self.__watcher_asserts = {}  # type: Dict[WatcherAssert]
        self.__main_loop = []  # type: List[LoopModuleMixin]

    def __module_by_class_or_class_name(self, class_name):
        if isinstance(class_name, str):
            try:
                module_name, class_name = class_name.rsplit('.', 1)
                module = __import__(module_name, globals(), locals(), [class_name], 0)
                clazz = getattr(module, class_name)
            except ImportError as e:
                raise InvalidModuleError("Class doesn't exist: " + str(class_name), e)
            except AttributeError as e:
                raise InvalidModuleError("Class doesn't exist: " + str(class_name), e)
        else:
            clazz = class_name
        # Validations
        assert clazz is not None, "module instance shouldn't be None"
        if not issubclass(clazz, Module):
            raise InvalidModuleError('Module must implement healthcheckbot.common.model.Module class')
        return clazz

    def __bucket_for_module_instance(self, clazz: Type) -> ModuleType:
        if issubclass(clazz, TriggerModule):
            return ModuleType(self.__triggers, 'trigger', TriggerModule)
        elif issubclass(clazz, OutputModule):
            return ModuleType(self.__outputs, 'output', OutputModule)
        elif issubclass(clazz, WatcherModule):
            return ModuleType(self.__watchers, 'watcher', WatcherModule)
        elif issubclass(clazz, WatcherAssert):
            return ModuleType(self.__watcher_asserts, 'assert', WatcherAssert)
        else:
            raise InvalidModuleError('Unknown module type: ' + clazz.__name__)

    def get_trigger_by_name(self, name) -> TriggerModule:
        return self.__triggers.get(name)

    def get_output_by_name(self, name):
        return self.__triggers.get(name)

    def register_module_instance(self, instance_name, class_name) -> Module:
        try:
            module_class = self.__module_by_class_or_class_name(class_name)
            module_type = self.__bucket_for_module_instance(module_class)
            module_inst = module_class(self)
            try:
                module_inst.on_initialized()
            except Exception as e:
                raise InvalidModuleError("Error during initialization", e)
            module_type.bucket[instance_name] = module_inst
            self.__logger.info(
                'Instantiated {} "{}"({})'.format(module_type.name, instance_name, module_class.__name__))
            if isinstance(module_inst, LoopModuleMixin):
                self.__main_loop.append(module_inst)
                self.__logger.info(
                    'Module "{}"({}) registered in application loop'.format(instance_name, module_class.__name__))
            module_inst.name = instance_name
            return module_inst
        except InvalidModuleError as e:
            raise InvalidModuleError('Unable to register module {}: '.format(class_name) + e.message, e)

    def get_instance_settings(self) -> InstanceSettings:
        return self.__instance_settings

    def main_loop(self):
        while not self.__terminating:
            for module in self.__main_loop:
                try:
                    module.step()
                except Exception as e:
                    self.__logger.error("Error in during main loop execution: " + str(e))
            time.sleep(0.01)

    def run_watcher(self, watcher: WatcherModule, trigger: TriggerModule) -> WatcherResult:
        reporter = ValidationReporter(watcher, trigger)
        try:
            state = watcher.obtain_state(trigger)
        except Exception as e:
            raise WatcherRuntimeError(
                'Watcher "{}" was unable to finish obtain state cycle: {}'.format(watcher.name, str(e)), e)
        if watcher.enable_assertions:
            try:
                watcher.do_assertions(state, reporter)
                for assertion in watcher.custom_assertions:
                    try:
                        assertion.do_assert(state, reporter, assertion.name)
                    except Exception as e:
                        reporter.error(assertion.name, 'Unexpected error: ' + str(e))

            except Exception as e:
                raise WatcherRuntimeError(
                    'Watcher "{}" was unable to run assertions check: {}'.format(watcher.name, str(e)), e)
        try:
            serialized_state = watcher.serialize_state(state)
            watcher_result = WatcherResult(serialized_state, reporter.errors)
        except Exception as e:
            raise WatcherRuntimeError(
                'Watcher "{}" was unable serialize state: {}'.format(watcher.name, str(e)), e)
        self.deliver_watcher_result(watcher, watcher_result)
        return watcher_result

    def deliver_watcher_result(self, watcher: WatcherModule, watcher_result: WatcherResult):
        for output_name, output in self.__outputs.items():
            try:
                output.output(watcher, watcher_result)
            except Exception as e:
                raise OutputRuntimeError(
                    'Unable to deliver watcher result via output "{}": {}'.format(output.name, str(e)), e)

    def shutdown(self):
        self.__logger.info("Initiating shutdown process")
        self.__terminating = True
        for bucket in (self.__watchers, self.__triggers, self.__outputs, self.__watcher_asserts):
            for name, module in bucket.items():
                try:
                    module.on_before_destroyed()
                    del module
                except Exception as e:
                    self.__logger.error(
                        "Error while destroying module {}({}): {}".format(name, module.__class__.__name__, e))
        self.__logger.info("Destroyed modules")
